from .kmer_database import KmerDB, KmerBuilder, FastaChunkReader

# from ctypes import *
import os
import inspect
loaded=False
try:
    import dna_jellyfish as jf
    loaded=True
except ImportError:
    pass


class JellyfishSDB(KmerDB):
    def __init__(self, kmer_size):
        if not loaded:
            raise ImportError("Jellyfish not loaded")
        self.kmer_size = kmer_size
        #self.db  = kmerGWAS.KmerGWAS_database(kmer_size) 
        self.db = None
        self.builder = JellyfishBuilder(kmer_size)
 
    def builder(self):
        return self.builder

    def load_from_fasta(self, filename, buffer_size=10000):
        # fasta = pysam.FastaFile(filename)
        fasta_iter = FastaChunkReader(
            filename, chunk_size=buffer_size, kmer_size=self.kmer_size
        )        
        for chunk in fasta_iter:
            self.db.add_kmers(chunk['seq'])
        self.db.sort_unique()

    def __getitem__(self, index):
        return self.db[index]

    def __len__(self):
        raise NotImplementedError

    def __contains__(self, key):
        key.canonicalize()
        return self.db[key] > 0

    def load(self, filename):
        self.db = jf.QueryMerFile(filename)


class JellyfishBuilder(KmerBuilder):
    def __init__(self, kmer_size):
        if not loaded:
            raise ImportError("Jellyfish not loaded")
        self.kmer_size = kmer_size

    def string_to_kmer(self, sequence):
        binary_kmer = jf.MerDNA(sequence)
        return binary_kmer

    def kmer_to_string(self, binary_kmer):
        return str(binary_kmer)

    def compare(self, a, b):
        ka = self.string_to_kmer(a)
        kb = self.string_to_kmer(b)
        return self._builder.compare(ka, kb)



import os
import sys
#import pathlib
#add_path = os.path.dirname(os.path.abspath(__file__))+ "/.." 
#print(add_path)
#sys.path.insert(0, add_path)

import logging
import operator
import unittest
from IBSpy import JellyfishSDB, FastaChunkReader 

loaded=False
try:
    import dna_jellyfish as jf
    loaded=True
except ImportError:
    pass

class TestJellyfish(unittest.TestCase):

	logger = logging.getLogger("test")
	logger.setLevel(logging.DEBUG)

	def setUp(self):
		self.data_path="./tests/data/"

			
	# def test_compare_kmers(self):
	# 	kmer_builder = KmerGWASDBBuilder(31)
	# 	tests = [
	# 	'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
	# 	'TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT',
	# 	'GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG',
	# 	'CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC',
	# 	'AAAAAAAAAGGGAAAGGGAAGAGGGAAAGGG']
	# 	self.assertEqual(kmer_builder.compare(tests[0], tests[1]),  0)
	# 	self.assertEqual(kmer_builder.compare(tests[2], tests[1]),  1)
	# 	self.assertEqual(kmer_builder.compare(tests[1], tests[2]), -1)

	def run_db_tests(self, kmerdb):
		references=['jagger', 'arinalrfor', 'julius', 'lancer', 'landmark', 'mace', 'norin61', 'spelta', 'stanley', 'sy_mattis']
		expected = [
		[2970, 2201], [2939, 2170], [2970, 2201], [2609],
		[2970, 2201], [2970, 2201], [2970, 2201], [2970, 2201],
		[2879, 2176], [2970, 2201]
		]
		i = 0
		for r in references:	
			path = self.data_path + "/test4B." + r + ".fa"
			windows = kmerdb.kmers_in_windows(path, window_size=3000)
			obs =  map(lambda w: w.observed_kmers, windows)
			j = 0
			for o in obs:
				self.assertEqual(expected[i][j], o)
				j += 1
			i += 1



	def test_jellyfish_db(self):
		if not loaded:
			print("JellyFish python module not installed", file=sys.stderr)
		else:
			kmerdb = JellyfishSDB(31)
			kmerdb.load(self.data_path + "/test4B.jagger.fa.k31.jf")
			self.run_db_tests(kmerdb)


if __name__ == '__main__':
	unittest.main()


cdef extern from "stdint.h":
	ctypedef unsigned long long uint64_t
	ctypedef unsigned short uint8_t

cdef extern from "kmer_general.h":
	ctypedef uint64_t kmerGWAS_kmer

	kmerGWAS_kmer kmer2bits(char * k, uint8_t kmer_size)
	char * kmer_string_alloc(const uint8_t kmer_size)
	void kmer_string_free(char * res)
	void bits2kmer31(kmerGWAS_kmer kmer, uint8_t kmer_size, char * res)
	int kmer_compare (kmerGWAS_kmer a, kmerGWAS_kmer b)

cdef extern from "kmer_db.h":
	struct KmerGwasTable:
		pass

	KmerGwasTable * kmer_gwas_table_new(uint8_t kmer_size)
	void kmer_gwas_table_free(KmerGwasTable * * kgt)
	void kmer_gwas_table_add_kmers_from_string(char * sequence, KmerGwasTable * kgt)
	kmerGWAS_kmer kmer_gwas_table_get(uint64_t index, KmerGwasTable * kgt)
	uint64_t kmer_gwas_sort_and_filter_unique(KmerGwasTable * kgt)
	kmerGWAS_kmer * kmer_gwas_table_find(kmerGWAS_kmer kmer, KmerGwasTable * kgt)
	void kmer_gwas_table_mmap_read(char * file, KmerGwasTable * kgt )
	void kmer_gwas_table_save(char * filename, KmerGwasTable * kgt)




from abc import ABC, abstractmethod, abstractproperty
from pyfaidx import Fasta
from functools import reduce 

class WindowStats:
    __dict__ = ['seqname', 'start', 'end','total_kmers', 
    'observed_kmers', 'variations', 'kmer_distance']

    def header(self, sep="\t"):
        return sep.join(vars(self))

    def csv(self, sep="\t"):
        vs = map(lambda w: getattr(self, w), vars(self))
        return sep.join(map(str, vs))

class KmerDB(ABC):
    
    @abstractproperty
    def builder(self):
        return None

    @abstractmethod
    def __contains__(self, key):
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    def count_kmers_from_sequence(self, kmers): 
        f = filter(lambda k: k in self, kmers)
        return reduce(lambda x, y: x + 1 , f, 0)

    def kmers_stats_from_sequence(self, kmers): 
        stats = WindowStats()
        stats.observed_kmers = 0
        stats.variations     = 0
        stats.kmer_distance  = 0
        stats.total_kmers = len(kmers)
        last_present = True
        gap_size = 0
        i = 0
        for k in kmers: 
            i += 1
            present = k in self 
            if not present:
                gap_size += 1
            else:
                stats.observed_kmers += 1
                if gap_size > 0:
                    stats.variations += 1
                    kmer_distance = gap_size - (self.kmer_size -1)
                    #print(str(i) + ": " + str(gap_size) + ": "+ str(kmer_distance))
                    if kmer_distance <= 0:
                        kmer_distance = abs(kmer_distance + 1)
                    #print(str(i) + ": " + str(gap_size) + ": "+ str(kmer_distance))
                    stats.kmer_distance += kmer_distance
                gap_size = 0;

            last_present = present
        #print(".....")
        if gap_size > 0:
            stats.variations += 1
            kmer_distance = gap_size - (self.kmer_size -1)
            if kmer_distance <= 0:
                kmer_distance = abs(kmer_distance + 1)
            stats.kmer_distance += kmer_distance
        return stats

    def kmers_in_windows(self, path, window_size=1000, chunk=0, total_chunks=1):
        def window_summary(seq):
            kmers = self.builder.sequence_to_kmers(seq['seq'], convert=True)
            stats = self.kmers_stats_from_sequence(kmers)
            stats.seqname = seq['seqname']
            stats.start   = seq['start']
            stats.end     = seq['end']
            return stats
        fasta_iter = FastaChunkReader(path, chunk_size = window_size, 
            kmer_size=self.kmer_size, chunk=chunk, total_chunks=total_chunks)
        return map(window_summary, fasta_iter )

class KmerBuilder(ABC):
    def __init__(self, kmer_size):
        self.kmer_size = kmer_size

    @abstractmethod
    def string_to_kmer(self, sequence):
        raise NotImplementedError

    @abstractmethod
    def kmer_to_string(self, sequence):
        raise NotImplementedError

    def sequence_to_kmers(self, sequence, filter_ambiguity=True, convert=False):
        ret = [None] * (len(sequence) - self.kmer_size + 1)
        sequence = sequence.upper()
        for start in range(0, len(sequence) - self.kmer_size + 1):
            end = start + self.kmer_size
            kmer = sequence[start:end]
            count = kmer.count("A")
            count += kmer.count("T")
            count += kmer.count("G")
            count += kmer.count("C")
            if filter_ambiguity and count != self.kmer_size:
                continue
            ret[start] = self.string_to_kmer(kmer) if convert else kmer 
        return [i for i in ret if i]


class FastaChunkReader:
    def __init__(self, filename, chunk_size=10000, kmer_size=31, 
        chunk=0, total_chunks=1):
        self.fasta         = Fasta(filename)
        self.current_ref   = 0
        self.current_start = 0
        self.chunk_size    = chunk_size
        self.kmer_size     = kmer_size
        self.seqnames      = list(self.fasta.keys())
        self.chunk         = chunk
        self.total_chunks  = total_chunks

        

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.seqnames) == self.current_ref:
            self.fasta.close()
            raise StopIteration
        seqname = self.seqnames[self.current_ref]
        start = self.current_start
        end = start + self.chunk_size
        self.current_start = end - self.kmer_size
        if end >= len(self.fasta[seqname]):
            self.current_start = 0
            self.current_ref += 1
            end = len(self.fasta[seqname])
        return {
            "seqname": seqname,
            "start": start,
            "end": end,
            "seq": self.fasta[seqname][start:end].seq,
        }



import argparse
import gzip
from sys import stdout
from IBSpy import KmerGWASDBBuilder, KmerGWASDB, FastaChunkReader, JellyfishSDB, JellyfishBuilder 

def window_count(args):
	window_size = 1000
	kmers_path  = "./tests/data/test4B.jagger.kmerGWAS_k31"
	reference  = "./tests/data/test4B.stanley.fa"
	kmer_size   = 31

	kmerdb = None

	if(args.database_format=="kmerGWAS"):
		kmerdb = KmerGWASDB(args.kmer_size)
	if(args.database_format=="jellyfish"):
		kmerdb = JellyfishSDB(args.kmer_size)
	
	kmerdb.load(args.database)
	windows = kmerdb.kmers_in_windows(args.reference, window_size=args.window_size)
	printed = False
	out = stdout

	if(args.compress and args.output is not None) :
		out = gzip.open(args.output + ".gz", 'w') 

	if(not args.compress and args.output is not None) :
		out = open(args.output , 'w') 

	for w in windows:
		if not printed:
			line = w.header()
			line += "\n"
			if args.compress:
				line=line.encode()
			out.write(line)
			printed = True
		line = w.csv()
		line += "\n"  
		if args.compress:
			line=line.encode()
		out.write(line)

	if(args.output is not None):
		out.close()


def main():
	parser = argparse.ArgumentParser()
	parser.add_argument("-w", "--window_size", default=500000,
		help="window size to analyze", type=int)
	parser.add_argument("-k", "--kmer_size", default=31,
		help="Kmer size of the database", type=int)
	parser.add_argument("-d", "--database", default="./tests/data/test4B.jagger.kmerGWAS_k31",
		help="Kmer database")
	parser.add_argument("-r", "--reference", default="./tests/data/test4B.stanley.fa",
		help="The reference with the position of the kmers")
	parser.add_argument("-z", "--compress", default=False, action="store_true", 
		help="When an ouput file is present, it is compressed as .gz")
	parser.add_argument("-o", "--output", default=None, 
		help="Output file. If missing, the ouptut is sent to stdout")
	parser.add_argument("-f", "--database_format", default="kmerGWAS", choices=["kmerGWAS", "jellyfish"],
		help="Database format (kmerGWAS, jellyfish)")
	args = parser.parse_args()
	window_count(args)

if __name__ == '__main__':
	main()

	# args = {
	# window_size : 1000000,
	# kmers_path  : "./tests/data/test4B.jagger.kmerGWAS_k31",
	# reference   : "./tests/data/test4B.stanley.fa",
	# kmer_size   : 31,
	# compress    : True,
	# output      : None
	# }

	#We need to add the different databases

	

	
	#parser = argparse.ArgumentParser()
	#parser.parse_args()

	


import setuptools
from distutils.core import setup, Extension
from Cython.Build import cythonize
import unittest

def unit_tests():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite

# See https://stackoverflow.com/a/51272967/5188860
module1 = Extension('kmerGWAS',
                   #sources = ['IBSpy/kmerGWAS/kmer_gwas.pyx']
                     sources = ['IBSpy/kmerGWAS/nucleotide.c',
                                'IBSpy/kmerGWAS/kmer_gwas.pyx',
                                'IBSpy/kmerGWAS/kmer_general.c',
                                'IBSpy/kmerGWAS/kmer_db.c']
                               )

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="IBSpy", # Replace with your own username
    version="0.0.9",
    author="Ricardo H. Ramirez-Gonzalez",
    author_email="ricardo.ramirez-gonzalez@jic.ac.uk",
    description="A package to detect IBS regions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Uauy-Lab/IBSpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    test_suite='setup.unit_tests',
    #ext_modules=[module1]
    ext_modules = cythonize(
        [module1],
        compiler_directives={'language_level': "3"}),
    entry_points={  # Optional
        'console_scripts': [
            'IBSpy_wcount=IBSpy.IBSpy_window_count:main',
            'IBSpy=IBSpy:main'
        ],
    }
)
